@@ -4,23 +4,22 @@
  * This file is part of GtkSourceView
  *
  * Copyright (C) 1998, 1999 Alex Roberts, Evan Lawrence
- * Copyright (C) 2000, 2001 Chema Celorio, Paolo Maggi 
- * Copyright (C) 2002-2005  Paolo Maggi 
+ * Copyright (C) 2000, 2001 Chema Celorio, Paolo Maggi
+ * Copyright (C) 2002-2005  Paolo Maggi
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
- * This program is distributed in the hope that it will be useful,
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, 
- * Boston, MA 02111-1307, USA.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #ifdef HAVE_CONFIG_H
@@ -35,8 +34,18 @@
 #include "gtksourceview-marshal.h"
 
 
-#define DEFAULT_MAX_UNDO_LEVELS		25
+#define DEFAULT_MAX_UNDO_LEVELS		-1
 
+/*
+ * The old code which used a GSList and g_slist_nth_element
+ * was way too slow for many operations (search/replace, hit Ctrl-Z,
+ * see gedit freezes). GSList was replaced with a GPtrArray with
+ * minimal changes to the code: to avoid insertions at the beginning
+ * of the array it uses array beginning as the list end and vice versa;
+ * hence bunch of ugly action_list_* functions.
+ * FIXME: so, somebody please rewrite this stuff using some nice data
+ * structures or something.
+ */
 
 typedef struct _GtkSourceUndoAction  			GtkSourceUndoAction;
 typedef struct _GtkSourceUndoInsertAction		GtkSourceUndoInsertAction;
@@ -47,14 +56,14 @@ typedef enum {
 	GTK_SOURCE_UNDO_ACTION_DELETE
 } GtkSourceUndoActionType;
 
-/* 
+/*
  * We use offsets instead of GtkTextIters because the last ones
  * require to much memory in this context without giving us any advantage.
- */ 
+ */
 
 struct _GtkSourceUndoInsertAction
 {
-	gint   pos; 
+	gint   pos;
 	gchar *text;
 	gint   length;
 	gint   chars;
@@ -71,7 +80,7 @@ struct _GtkSourceUndoDeleteAction
 struct _GtkSourceUndoAction
 {
 	GtkSourceUndoActionType action_type;
-	
+
 	union {
 		GtkSourceUndoInsertAction  insert;
 		GtkSourceUndoDeleteAction  delete;
@@ -83,7 +92,7 @@ struct _GtkSourceUndoAction
 	guint mergeable : 1;
 
 	/* It is TRUE whether the action is marked as "modified".
-	 * An action is marked as "modified" if it changed the 
+	 * An action is marked as "modified" if it changed the
 	 * state of the buffer from "not modified" to "modified". Only the first
 	 * action of a group can be marked as modified.
 	 * There can be a single action marked as "modified" in the actions list.
@@ -97,29 +106,29 @@ struct _GtkSourceUndoAction
 struct _GtkSourceUndoManagerPrivate
 {
 	GtkTextBuffer	*document;
-	
-	GList*		 actions;
-	gint 		 next_redo;	
+
+	GPtrArray	*actions;
+	gint 		 next_redo;
 
 	gint 		 actions_in_current_group;
-	
+
 	gint		 running_not_undoable_actions;
 
 	gint		 num_of_groups;
 
 	gint		 max_undo_levels;
-	
+
 	guint	 	 can_undo : 1;
 	guint		 can_redo : 1;
-	
+
 	/* It is TRUE whether, while undoing an action of the current group (with order_in_group > 1),
 	 * the state of the buffer changed from "not modified" to "modified".
 	 */
-	guint	 	 modified_undoing_group : 1;	
+	guint	 	 modified_undoing_group : 1;
 
 	/* Pointer to the action (in the action list) marked as "modified".
-	 * It is NULL when no action is marked as "modified". 
-	 * It is INVALID when the action marked as "modified" has been removed 
+	 * It is NULL when no action is marked as "modified".
+	 * It is INVALID when the action marked as "modified" has been removed
 	 * from the action list (freeing the list or resizing it) */
 	GtkSourceUndoAction *modified_action;
 };
@@ -130,86 +139,56 @@ enum {
 	LAST_SIGNAL
 };
 
+G_DEFINE_TYPE (GtkSourceUndoManager, gtk_source_undo_manager, G_TYPE_OBJECT)
+
 static void gtk_source_undo_manager_class_init 			(GtkSourceUndoManagerClass 	*klass);
 static void gtk_source_undo_manager_init 			(GtkSourceUndoManager 		*um);
 static void gtk_source_undo_manager_finalize 			(GObject 			*object);
 
-static void gtk_source_undo_manager_insert_text_handler 	(GtkTextBuffer 			*buffer, 
+static void gtk_source_undo_manager_insert_text_handler 	(GtkTextBuffer 			*buffer,
 							 	 GtkTextIter 			*pos,
-		                             		 	 const 	gchar 			*text, 
-							 	 gint 				 length, 
+		                             		 	 const 	gchar 			*text,
+							 	 gint 				 length,
 							 	 GtkSourceUndoManager 		*um);
-static void gtk_source_undo_manager_delete_range_handler 	(GtkTextBuffer 			*buffer, 
+static void gtk_source_undo_manager_delete_range_handler 	(GtkTextBuffer 			*buffer,
 							 	 GtkTextIter 			*start,
                         		      		 	 GtkTextIter 			*end,
 							 	 GtkSourceUndoManager 		*um);
-static void gtk_source_undo_manager_begin_user_action_handler 	(GtkTextBuffer 			*buffer, 
+static void gtk_source_undo_manager_begin_user_action_handler 	(GtkTextBuffer 			*buffer,
 								 GtkSourceUndoManager 		*um);
 static void gtk_source_undo_manager_modified_changed_handler	(GtkTextBuffer                  *buffer,
 								 GtkSourceUndoManager           *um);
 
 static void gtk_source_undo_manager_free_action_list 		(GtkSourceUndoManager 		*um);
 
-static void gtk_source_undo_manager_add_action 			(GtkSourceUndoManager 		*um, 
+static void gtk_source_undo_manager_add_action 			(GtkSourceUndoManager 		*um,
 		                                         	 const GtkSourceUndoAction 	*undo_action);
-static void gtk_source_undo_manager_free_first_n_actions 	(GtkSourceUndoManager 		*um, 
+static void gtk_source_undo_manager_free_first_n_actions 	(GtkSourceUndoManager 		*um,
 								 gint 				 n);
 static void gtk_source_undo_manager_check_list_size 		(GtkSourceUndoManager 		*um);
 
-static gboolean gtk_source_undo_manager_merge_action 		(GtkSourceUndoManager 		*um, 
+static gboolean gtk_source_undo_manager_merge_action 		(GtkSourceUndoManager 		*um,
 		                                        	 const GtkSourceUndoAction 	*undo_action);
 
-static GObjectClass 	*parent_class 				= NULL;
 static guint 		undo_manager_signals [LAST_SIGNAL] 	= { 0 };
 
-GType
-gtk_source_undo_manager_get_type (void)
-{
-	static GType undo_manager_type = 0;
-
-  	if (undo_manager_type == 0)
-    	{
-      		static const GTypeInfo our_info =
-      		{
-        		sizeof (GtkSourceUndoManagerClass),
-        		NULL,		/* base_init */
-        		NULL,		/* base_finalize */
-        		(GClassInitFunc) gtk_source_undo_manager_class_init,
-        		NULL,           /* class_finalize */
-        		NULL,           /* class_data */
-        		sizeof (GtkSourceUndoManager),
-        		0,              /* n_preallocs */
-        		(GInstanceInitFunc) gtk_source_undo_manager_init
-      		};
-
-      		undo_manager_type = g_type_register_static (G_TYPE_OBJECT,
-                					    "GtkSourceUndoManager",
-							    &our_info,
-							    0);
-    	}
-
-	return undo_manager_type;
-}
-
 static void
 gtk_source_undo_manager_class_init (GtkSourceUndoManagerClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
-  	parent_class = g_type_class_peek_parent (klass);
-
   	object_class->finalize = gtk_source_undo_manager_finalize;
 
         klass->can_undo 	= NULL;
 	klass->can_redo 	= NULL;
-	
+
 	undo_manager_signals[CAN_UNDO] =
    		g_signal_new ("can_undo",
 			      G_OBJECT_CLASS_TYPE (object_class),
 			      G_SIGNAL_RUN_LAST,
 			      G_STRUCT_OFFSET (GtkSourceUndoManagerClass, can_undo),
 			      NULL, NULL,
-			      gtksourceview_marshal_VOID__BOOLEAN,
+			      _gtksourceview_marshal_VOID__BOOLEAN,
 			      G_TYPE_NONE,
 			      1,
 			      G_TYPE_BOOLEAN);
@@ -220,19 +199,21 @@ gtk_source_undo_manager_class_init (GtkSourceUndoManagerClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      G_STRUCT_OFFSET (GtkSourceUndoManagerClass, can_redo),
 			      NULL, NULL,
-			      gtksourceview_marshal_VOID__BOOLEAN,
+			      _gtksourceview_marshal_VOID__BOOLEAN,
 			      G_TYPE_NONE,
 			      1,
 			      G_TYPE_BOOLEAN);
 
+	g_type_class_add_private (object_class, sizeof(GtkSourceUndoManagerPrivate));
 }
 
 static void
 gtk_source_undo_manager_init (GtkSourceUndoManager *um)
 {
-	um->priv = g_new0 (GtkSourceUndoManagerPrivate, 1);
+	um->priv = G_TYPE_INSTANCE_GET_PRIVATE (um, GTK_TYPE_SOURCE_UNDO_MANAGER,
+						GtkSourceUndoManagerPrivate);
 
-	um->priv->actions = NULL;
+	um->priv->actions = g_ptr_array_new ();
 	um->priv->next_redo = 0;
 
 	um->priv->can_undo = FALSE;
@@ -256,31 +237,27 @@ gtk_source_undo_manager_finalize (GObject *object)
 
 	g_return_if_fail (object != NULL);
 	g_return_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (object));
-	
+
    	um = GTK_SOURCE_UNDO_MANAGER (object);
 
 	g_return_if_fail (um->priv != NULL);
 
-	if (um->priv->actions != NULL)
-	{
-		gtk_source_undo_manager_free_action_list (um);
-	}
+	gtk_source_undo_manager_free_action_list (um);
+	g_ptr_array_free (um->priv->actions, TRUE);
 
 	g_signal_handlers_disconnect_by_func (G_OBJECT (um->priv->document),
-			  G_CALLBACK (gtk_source_undo_manager_delete_range_handler), 
+			  G_CALLBACK (gtk_source_undo_manager_delete_range_handler),
 			  um);
 
 	g_signal_handlers_disconnect_by_func (G_OBJECT (um->priv->document),
-			  G_CALLBACK (gtk_source_undo_manager_insert_text_handler), 
+			  G_CALLBACK (gtk_source_undo_manager_insert_text_handler),
 			  um);
-	
+
 	g_signal_handlers_disconnect_by_func (G_OBJECT (um->priv->document),
-			  G_CALLBACK (gtk_source_undo_manager_begin_user_action_handler), 
+			  G_CALLBACK (gtk_source_undo_manager_begin_user_action_handler),
 			  um);
 
-	g_free (um->priv);
-
-	G_OBJECT_CLASS (parent_class)->finalize (object);
+	G_OBJECT_CLASS (gtk_source_undo_manager_parent_class)->finalize (object);
 }
 
 GtkSourceUndoManager*
@@ -288,21 +265,21 @@ gtk_source_undo_manager_new (GtkTextBuffer* buffer)
 {
  	GtkSourceUndoManager *um;
 
-	um = GTK_SOURCE_UNDO_MANAGER (g_object_new (GTK_SOURCE_TYPE_UNDO_MANAGER, NULL));
+	um = GTK_SOURCE_UNDO_MANAGER (g_object_new (GTK_TYPE_SOURCE_UNDO_MANAGER, NULL));
 
 	g_return_val_if_fail (um->priv != NULL, NULL);
   	um->priv->document = buffer;
 
 	g_signal_connect (G_OBJECT (buffer), "insert_text",
-			  G_CALLBACK (gtk_source_undo_manager_insert_text_handler), 
+			  G_CALLBACK (gtk_source_undo_manager_insert_text_handler),
 			  um);
 
 	g_signal_connect (G_OBJECT (buffer), "delete_range",
-			  G_CALLBACK (gtk_source_undo_manager_delete_range_handler), 
+			  G_CALLBACK (gtk_source_undo_manager_delete_range_handler),
 			  um);
 
 	g_signal_connect (G_OBJECT (buffer), "begin_user_action",
-			  G_CALLBACK (gtk_source_undo_manager_begin_user_action_handler), 
+			  G_CALLBACK (gtk_source_undo_manager_begin_user_action_handler),
 			  um);
 
 	g_signal_connect (G_OBJECT (buffer), "modified_changed",
@@ -311,7 +288,7 @@ gtk_source_undo_manager_new (GtkTextBuffer* buffer)
 	return um;
 }
 
-void 
+void
 gtk_source_undo_manager_begin_not_undoable_action (GtkSourceUndoManager *um)
 {
 	g_return_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um));
@@ -320,18 +297,18 @@ gtk_source_undo_manager_begin_not_undoable_action (GtkSourceUndoManager *um)
 	++um->priv->running_not_undoable_actions;
 }
 
-static void 
+static void
 gtk_source_undo_manager_end_not_undoable_action_internal (GtkSourceUndoManager *um)
 {
 	g_return_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um));
 	g_return_if_fail (um->priv != NULL);
 
 	g_return_if_fail (um->priv->running_not_undoable_actions > 0);
-	
+
 	--um->priv->running_not_undoable_actions;
 }
 
-void 
+void
 gtk_source_undo_manager_end_not_undoable_action (GtkSourceUndoManager *um)
 {
 	g_return_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um));
@@ -340,32 +317,31 @@ gtk_source_undo_manager_end_not_undoable_action (GtkSourceUndoManager *um)
 	gtk_source_undo_manager_end_not_undoable_action_internal (um);
 
 	if (um->priv->running_not_undoable_actions == 0)
-	{	
+	{
 		gtk_source_undo_manager_free_action_list (um);
-	
-		um->priv->next_redo = -1;	
+
+		um->priv->next_redo = -1;
 
 		if (um->priv->can_undo)
 		{
 			um->priv->can_undo = FALSE;
-			g_signal_emit (G_OBJECT (um), 
-				       undo_manager_signals [CAN_UNDO], 
-				       0, 
+			g_signal_emit (G_OBJECT (um),
+				       undo_manager_signals [CAN_UNDO],
+				       0,
 				       FALSE);
 		}
 
 		if (um->priv->can_redo)
 		{
 			um->priv->can_redo = FALSE;
-			g_signal_emit (G_OBJECT (um), 
-				       undo_manager_signals [CAN_REDO], 
-				       0, 
+			g_signal_emit (G_OBJECT (um),
+				       undo_manager_signals [CAN_REDO],
+				       0,
 				       FALSE);
 		}
 	}
 }
 
-
 gboolean
 gtk_source_undo_manager_can_undo (const GtkSourceUndoManager *um)
 {
@@ -375,7 +351,7 @@ gtk_source_undo_manager_can_undo (const GtkSourceUndoManager *um)
 	return um->priv->can_undo;
 }
 
-gboolean 
+gboolean
 gtk_source_undo_manager_can_redo (const GtkSourceUndoManager *um)
 {
 	g_return_val_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um), FALSE);
@@ -388,22 +364,22 @@ static void
 set_cursor (GtkTextBuffer *buffer, gint cursor)
 {
 	GtkTextIter iter;
-	
+
 	/* Place the cursor at the requested position */
 	gtk_text_buffer_get_iter_at_offset (buffer, &iter, cursor);
 	gtk_text_buffer_place_cursor (buffer, &iter);
 }
 
-static void 
+static void
 insert_text (GtkTextBuffer *buffer, gint pos, const gchar *text, gint len)
 {
 	GtkTextIter iter;
-	
+
 	gtk_text_buffer_get_iter_at_offset (buffer, &iter, pos);
 	gtk_text_buffer_insert (buffer, &iter, text, len);
 }
 
-static void 
+static void
 delete_text (GtkTextBuffer *buffer, gint start, gint end)
 {
 	GtkTextIter start_iter;
@@ -435,23 +411,60 @@ get_chars (GtkTextBuffer *buffer, gint start, gint end)
 	return gtk_text_buffer_get_slice (buffer, &start_iter, &end_iter, TRUE);
 }
 
-void 
+static GtkSourceUndoAction *
+action_list_nth_data (GPtrArray *array,
+		      gint       n)
+{
+	if (n < 0 || n >= (gint)array->len)
+		return NULL;
+	else
+		return array->pdata[array->len - 1 - n];
+}
+
+static void
+action_list_prepend (GPtrArray           *array,
+		     GtkSourceUndoAction *action)
+{
+	g_ptr_array_add (array, action);
+}
+
+static GtkSourceUndoAction *
+action_list_last_data (GPtrArray *array)
+{
+	if (array->len != 0)
+		return array->pdata[0];
+	else
+		return NULL;
+}
+
+static void
+action_list_delete_last (GPtrArray *array)
+{
+	if (array->len != 0)
+	{
+		memmove (&array->pdata[0], &array->pdata[1], (array->len - 1)*sizeof (gpointer));
+		g_ptr_array_set_size (array, array->len - 1);
+	}
+}
+
+void
 gtk_source_undo_manager_undo (GtkSourceUndoManager *um)
 {
 	GtkSourceUndoAction *undo_action;
 	gboolean modified = FALSE;
+	gint cursor_pos = -1;
 
 	g_return_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um));
 	g_return_if_fail (um->priv != NULL);
 	g_return_if_fail (um->priv->can_undo);
-	
+
 	um->priv->modified_undoing_group = FALSE;
 
 	gtk_source_undo_manager_begin_not_undoable_action (um);
 
 	do
-	{	
-		undo_action = g_list_nth_data (um->priv->actions, um->priv->next_redo + 1);
+	{
+		undo_action = action_list_nth_data (um->priv->actions, um->priv->next_redo + 1);
 		g_return_if_fail (undo_action != NULL);
 
 		/* undo_action->modified can be TRUE only if undo_action->order_in_group <= 1 */
@@ -461,7 +474,7 @@ gtk_source_undo_manager_undo (GtkSourceUndoManager *um)
 		if (undo_action->order_in_group <= 1)
 		{
 			/* Set modified to TRUE only if the buffer did not change its state from
-			 * "not modified" to "modified" undoing an action (with order_in_group > 1) 
+			 * "not modified" to "modified" undoing an action (with order_in_group > 1)
 			 * in current group. */
 			modified = (undo_action->modified && !um->priv->modified_undoing_group);
 		}
@@ -470,32 +483,26 @@ gtk_source_undo_manager_undo (GtkSourceUndoManager *um)
 		{
 			case GTK_SOURCE_UNDO_ACTION_DELETE:
 				insert_text (
-					um->priv->document, 
-					undo_action->action.delete.start, 
+					um->priv->document,
+					undo_action->action.delete.start,
 					undo_action->action.delete.text,
 					strlen (undo_action->action.delete.text));
 
 				if (undo_action->action.delete.forward)
-					set_cursor (
-						um->priv->document, 
-						undo_action->action.delete.start);
+					cursor_pos = undo_action->action.delete.start;
 				else
-					set_cursor (
-						um->priv->document, 
-						undo_action->action.delete.end);
+					cursor_pos = undo_action->action.delete.end;
 
 				break;
-				
+
 			case GTK_SOURCE_UNDO_ACTION_INSERT:
 				delete_text (
-					um->priv->document, 
-					undo_action->action.insert.pos, 
-					undo_action->action.insert.pos + 
-						undo_action->action.insert.chars); 
+					um->priv->document,
+					undo_action->action.insert.pos,
+					undo_action->action.insert.pos +
+						undo_action->action.insert.chars);
 
-				set_cursor (
-					um->priv->document, 
-					undo_action->action.insert.pos);
+				cursor_pos = undo_action->action.insert.pos;
 				break;
 
 			default:
@@ -507,6 +514,9 @@ gtk_source_undo_manager_undo (GtkSourceUndoManager *um)
 
 	} while (undo_action->order_in_group > 1);
 
+	if (cursor_pos >= 0)
+		set_cursor (um->priv->document, cursor_pos);
+
 	if (modified)
 	{
 		--um->priv->next_redo;
@@ -515,39 +525,40 @@ gtk_source_undo_manager_undo (GtkSourceUndoManager *um)
 	}
 
 	gtk_source_undo_manager_end_not_undoable_action_internal (um);
-	
+
 	um->priv->modified_undoing_group = FALSE;
 
 	if (!um->priv->can_redo)
 	{
 		um->priv->can_redo = TRUE;
-		g_signal_emit (G_OBJECT (um), 
-			       undo_manager_signals [CAN_REDO], 
-			       0, 
+		g_signal_emit (G_OBJECT (um),
+			       undo_manager_signals [CAN_REDO],
+			       0,
 			       TRUE);
 	}
 
-	if (um->priv->next_redo >= (gint)(g_list_length (um->priv->actions) - 1))
+	if (um->priv->next_redo >= (gint)um->priv->actions->len - 1)
 	{
 		um->priv->can_undo = FALSE;
-		g_signal_emit (G_OBJECT (um), 
-			       undo_manager_signals [CAN_UNDO], 
-			       0, 
+		g_signal_emit (G_OBJECT (um),
+			       undo_manager_signals [CAN_UNDO],
+			       0,
 			       FALSE);
 	}
 }
 
-void 
+void
 gtk_source_undo_manager_redo (GtkSourceUndoManager *um)
 {
 	GtkSourceUndoAction *undo_action;
 	gboolean modified = FALSE;
+	gint cursor_pos = -1;
 
 	g_return_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um));
 	g_return_if_fail (um->priv != NULL);
 	g_return_if_fail (um->priv->can_redo);
-	
-	undo_action = g_list_nth_data (um->priv->actions, um->priv->next_redo);
+
+	undo_action = action_list_nth_data (um->priv->actions, um->priv->next_redo);
 	g_return_if_fail (undo_action != NULL);
 
 	gtk_source_undo_manager_begin_not_undoable_action (um);
@@ -561,29 +572,26 @@ gtk_source_undo_manager_redo (GtkSourceUndoManager *um)
 		}
 
 		--um->priv->next_redo;
-	
+
 		switch (undo_action->action_type)
 		{
 			case GTK_SOURCE_UNDO_ACTION_DELETE:
 				delete_text (
-					um->priv->document, 
-					undo_action->action.delete.start, 
-					undo_action->action.delete.end); 
-
-				set_cursor (
 					um->priv->document,
-					undo_action->action.delete.start);
+					undo_action->action.delete.start,
+					undo_action->action.delete.end);
+
+				cursor_pos = undo_action->action.delete.start;
 
 				break;
-				
+
 			case GTK_SOURCE_UNDO_ACTION_INSERT:
-				set_cursor (
-					um->priv->document,
-					undo_action->action.insert.pos);
+				cursor_pos = undo_action->action.insert.pos +
+						undo_action->action.insert.length;
 
 				insert_text (
-					um->priv->document, 
-					undo_action->action.insert.pos, 
+					um->priv->document,
+					undo_action->action.insert.pos,
 					undo_action->action.insert.text,
 					undo_action->action.insert.length);
 
@@ -598,10 +606,13 @@ gtk_source_undo_manager_redo (GtkSourceUndoManager *um)
 		if (um->priv->next_redo < 0)
 			undo_action = NULL;
 		else
-			undo_action = g_list_nth_data (um->priv->actions, um->priv->next_redo);
-			
+			undo_action = action_list_nth_data (um->priv->actions, um->priv->next_redo);
+
 	} while ((undo_action != NULL) && (undo_action->order_in_group > 1));
 
+	if (cursor_pos >= 0)
+		set_cursor (um->priv->document, cursor_pos);
+
 	if (modified)
 	{
 		++um->priv->next_redo;
@@ -622,61 +633,68 @@ gtk_source_undo_manager_redo (GtkSourceUndoManager *um)
 		um->priv->can_undo = TRUE;
 		g_signal_emit (G_OBJECT (um), undo_manager_signals [CAN_UNDO], 0, TRUE);
 	}
-
 }
 
-static void 
+static void
+gtk_source_undo_action_free (GtkSourceUndoAction *action)
+{
+	if (action == NULL)
+		return;
+
+	if (action->action_type == GTK_SOURCE_UNDO_ACTION_INSERT)
+		g_free (action->action.insert.text);
+	else if (action->action_type == GTK_SOURCE_UNDO_ACTION_DELETE)
+		g_free (action->action.delete.text);
+	else
+		g_return_if_reached ();
+
+	g_free (action);
+}
+
+static void
 gtk_source_undo_manager_free_action_list (GtkSourceUndoManager *um)
 {
-	gint n, len;
-	
-	g_return_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um));
-	g_return_if_fail (um->priv != NULL);
+	gint i;
 
-	if (um->priv->actions == NULL)
-		return;
-	
-	len = g_list_length (um->priv->actions);
-	
-	for (n = 0; n < len; n++)
+	for (i = (gint)um->priv->actions->len - 1; i >= 0; i--)
 	{
-		GtkSourceUndoAction *undo_action = 
-			(GtkSourceUndoAction *)(g_list_nth_data (um->priv->actions, n));
+		GtkSourceUndoAction *action = um->priv->actions->pdata[i];
 
-		if (undo_action->action_type == GTK_SOURCE_UNDO_ACTION_INSERT)
-			g_free (undo_action->action.insert.text);
-		else if (undo_action->action_type == GTK_SOURCE_UNDO_ACTION_DELETE)
-			g_free (undo_action->action.delete.text);
-		else
-			g_return_if_fail (FALSE);
-
-		if (undo_action->order_in_group == 1)
+		if (action->order_in_group == 1)
 			--um->priv->num_of_groups;
 
-		if (undo_action->modified)
+		if (action->modified)
 			um->priv->modified_action = INVALID;
 
-		g_free (undo_action);
+		gtk_source_undo_action_free (action);
 	}
 
-	g_list_free (um->priv->actions);
-	um->priv->actions = NULL;	
+	/* Some arbitrary limit, to avoid wasting space */
+	if (um->priv->actions->len > 2048)
+	{
+		g_ptr_array_free (um->priv->actions, TRUE);
+		um->priv->actions = g_ptr_array_new ();
+	}
+	else
+	{
+		g_ptr_array_set_size (um->priv->actions, 0);
+	}
 }
 
-static void 
-gtk_source_undo_manager_insert_text_handler (GtkTextBuffer 		*buffer, 
+static void
+gtk_source_undo_manager_insert_text_handler (GtkTextBuffer 		*buffer,
 					     GtkTextIter 		*pos,
-		                             const gchar 		*text, 
-					     gint 			 length, 
+		                             const gchar 		*text,
+					     gint 			 length,
 					     GtkSourceUndoManager 	*um)
 {
 	GtkSourceUndoAction undo_action;
-	
+
 	if (um->priv->running_not_undoable_actions > 0)
 		return;
 
 	g_return_if_fail (strlen (text) >= (guint)length);
-	
+
 	undo_action.action_type = GTK_SOURCE_UNDO_ACTION_INSERT;
 
 	undo_action.action.insert.pos    = gtk_text_iter_get_offset (pos);
@@ -695,15 +713,15 @@ gtk_source_undo_manager_insert_text_handler (GtkTextBuffer 		*buffer,
 	gtk_source_undo_manager_add_action (um, &undo_action);
 }
 
-static void 
-gtk_source_undo_manager_delete_range_handler (GtkTextBuffer 		*buffer, 
+static void
+gtk_source_undo_manager_delete_range_handler (GtkTextBuffer 		*buffer,
 					      GtkTextIter 		*start,
-                        		      GtkTextIter 		*end, 
+                        		      GtkTextIter 		*end,
 					      GtkSourceUndoManager 	*um)
 {
 	GtkSourceUndoAction undo_action;
 	GtkTextIter insert_iter;
-	
+
 	if (um->priv->running_not_undoable_actions > 0)
 		return;
 
@@ -734,14 +752,14 @@ gtk_source_undo_manager_delete_range_handler (GtkTextBuffer 		*buffer,
 		undo_action.mergeable = TRUE;
 
 	undo_action.modified = FALSE;
-	
+
 	gtk_source_undo_manager_add_action (um, &undo_action);
 
 	g_free (undo_action.action.delete.text);
 
 }
 
-static void 
+static void
 gtk_source_undo_manager_begin_user_action_handler (GtkTextBuffer *buffer, GtkSourceUndoManager *um)
 {
 	g_return_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um));
@@ -754,11 +772,11 @@ gtk_source_undo_manager_begin_user_action_handler (GtkTextBuffer *buffer, GtkSou
 }
 
 static void
-gtk_source_undo_manager_add_action (GtkSourceUndoManager 	*um, 
+gtk_source_undo_manager_add_action (GtkSourceUndoManager 	*um,
 				    const GtkSourceUndoAction 	*undo_action)
 {
 	GtkSourceUndoAction* action;
-	
+
 	if (um->priv->next_redo >= 0)
 	{
 		gtk_source_undo_manager_free_first_n_actions (um, um->priv->next_redo + 1);
@@ -774,22 +792,22 @@ gtk_source_undo_manager_add_action (GtkSourceUndoManager 	*um,
 		if (action->action_type == GTK_SOURCE_UNDO_ACTION_INSERT)
 			action->action.insert.text = g_strdup (undo_action->action.insert.text);
 		else if (action->action_type == GTK_SOURCE_UNDO_ACTION_DELETE)
-			action->action.delete.text = g_strdup (undo_action->action.delete.text); 
+			action->action.delete.text = g_strdup (undo_action->action.delete.text);
 		else
 		{
 			g_free (action);
 			g_return_if_reached ();
 		}
-		
+
 		++um->priv->actions_in_current_group;
 		action->order_in_group = um->priv->actions_in_current_group;
 
 		if (action->order_in_group == 1)
 			++um->priv->num_of_groups;
-	
-		um->priv->actions = g_list_prepend (um->priv->actions, action);
+
+		action_list_prepend (um->priv->actions, action);
 	}
-	
+
 	gtk_source_undo_manager_check_list_size (um);
 
 	if (!um->priv->can_undo)
@@ -805,121 +823,97 @@ gtk_source_undo_manager_add_action (GtkSourceUndoManager 	*um,
 	}
 }
 
-static void 
-gtk_source_undo_manager_free_first_n_actions (GtkSourceUndoManager	*um, 
+static void
+gtk_source_undo_manager_free_first_n_actions (GtkSourceUndoManager	*um,
 					      gint 			 n)
 {
 	gint i;
-	
-	g_return_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um));
-	g_return_if_fail (um->priv != NULL);
 
-	if (um->priv->actions == NULL)
+	if (um->priv->actions->len == 0)
 		return;
-	
+
 	for (i = 0; i < n; i++)
 	{
-		GtkSourceUndoAction *undo_action = 
-			(GtkSourceUndoAction *)(g_list_first (um->priv->actions)->data);
-	
-		if (undo_action->action_type == GTK_SOURCE_UNDO_ACTION_INSERT)
-			g_free (undo_action->action.insert.text);
-		else if (undo_action->action_type == GTK_SOURCE_UNDO_ACTION_DELETE)
-			g_free (undo_action->action.delete.text);
-		else
-			g_return_if_fail (FALSE);
+		GtkSourceUndoAction *action = um->priv->actions->pdata[um->priv->actions->len - 1];
 
-		if (undo_action->order_in_group == 1)
+		if (action->order_in_group == 1)
 			--um->priv->num_of_groups;
 
-		if (undo_action->modified)
+		if (action->modified)
 			um->priv->modified_action = INVALID;
-	
-		g_free (undo_action);
 
-		um->priv->actions = g_list_delete_link (um->priv->actions, um->priv->actions);
+		gtk_source_undo_action_free (action);
 
-		if (um->priv->actions == NULL) 
+		g_ptr_array_set_size (um->priv->actions, um->priv->actions->len - 1);
+
+		if (um->priv->actions->len == 0)
 			return;
 	}
 }
 
-static void 
+static void
 gtk_source_undo_manager_check_list_size (GtkSourceUndoManager *um)
 {
 	gint undo_levels;
-	
+
 	g_return_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um));
 	g_return_if_fail (um->priv != NULL);
-	
+
 	undo_levels = gtk_source_undo_manager_get_max_undo_levels (um);
-	
+
 	if (undo_levels < 1)
 		return;
 
 	if (um->priv->num_of_groups > undo_levels)
 	{
 		GtkSourceUndoAction *undo_action;
-		GList *last;
-		
-		last = g_list_last (um->priv->actions);
-		undo_action = (GtkSourceUndoAction*) last->data;
-			
+
+		undo_action = action_list_last_data (um->priv->actions);
+
 		do
 		{
-			GList *tmp;
-			
-			if (undo_action->action_type == GTK_SOURCE_UNDO_ACTION_INSERT)
-				g_free (undo_action->action.insert.text);
-			else if (undo_action->action_type == GTK_SOURCE_UNDO_ACTION_DELETE)
-				g_free (undo_action->action.delete.text);
-			else
-				g_return_if_fail (FALSE);
-
 			if (undo_action->order_in_group == 1)
 				--um->priv->num_of_groups;
 
 			if (undo_action->modified)
 				um->priv->modified_action = INVALID;
-			
-			g_free (undo_action);
 
-			tmp = g_list_previous (last);
-			um->priv->actions = g_list_delete_link (um->priv->actions, last);
-			last = tmp;
-			g_return_if_fail (last != NULL); 
+			gtk_source_undo_action_free (undo_action);
 
-			undo_action = (GtkSourceUndoAction*) last->data;
+			action_list_delete_last (um->priv->actions);
 
-		} while ((undo_action->order_in_group > 1) || 
+			undo_action = action_list_last_data (um->priv->actions);
+			g_return_if_fail (undo_action != NULL);
+
+		} while ((undo_action->order_in_group > 1) ||
 			 (um->priv->num_of_groups > undo_levels));
-	}	
+	}
 }
 
 /**
  * gtk_source_undo_manager_merge_action:
- * @um: a #GtkSourceUndoManager. 
+ * @um: a #GtkSourceUndoManager.
  * @undo_action: a #GtkSourceUndoAction.
- * 
+ *
  * This function tries to merge the undo action at the top of
  * the stack with a new undo action. So when we undo for example
  * typing, we can undo the whole word and not each letter by itself.
- * 
+ *
  * Return Value: %TRUE is merge was sucessful, %FALSE otherwise.
  **/
-static gboolean 
-gtk_source_undo_manager_merge_action (GtkSourceUndoManager 	*um, 
+static gboolean
+gtk_source_undo_manager_merge_action (GtkSourceUndoManager 	*um,
 				      const GtkSourceUndoAction *undo_action)
 {
 	GtkSourceUndoAction *last_action;
-	
+
 	g_return_val_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um), FALSE);
 	g_return_val_if_fail (um->priv != NULL, FALSE);
 
-	if (um->priv->actions == NULL)
+	if (um->priv->actions->len == 0)
 		return FALSE;
 
-	last_action = (GtkSourceUndoAction*) g_list_nth_data (um->priv->actions, 0);
+	last_action = action_list_nth_data (um->priv->actions, 0);
 
 	if (!last_action->mergeable)
 		return FALSE;
@@ -932,7 +926,7 @@ gtk_source_undo_manager_merge_action (GtkSourceUndoManager 	*um,
 	}
 
 	if (undo_action->action_type == GTK_SOURCE_UNDO_ACTION_DELETE)
-	{				
+	{
 		if ((last_action->action.delete.forward != undo_action->action.delete.forward) ||
 		    ((last_action->action.delete.start != undo_action->action.delete.start) &&
 		     (last_action->action.delete.start != undo_action->action.delete.end)))
@@ -940,14 +934,14 @@ gtk_source_undo_manager_merge_action (GtkSourceUndoManager 	*um,
 			last_action->mergeable = FALSE;
 			return FALSE;
 		}
-		
+
 		if (last_action->action.delete.start == undo_action->action.delete.start)
 		{
 			gchar *str;
-			
+
 #define L  (last_action->action.delete.end - last_action->action.delete.start - 1)
 #define g_utf8_get_char_at(p,i) g_utf8_get_char(g_utf8_offset_to_pointer((p),(i)))
-		
+
 			/* Deleted with the delete key */
 			if ((g_utf8_get_char (undo_action->action.delete.text) != ' ') &&
 			    (g_utf8_get_char (undo_action->action.delete.text) != '\t') &&
@@ -957,19 +951,19 @@ gtk_source_undo_manager_merge_action (GtkSourceUndoManager 	*um,
 				last_action->mergeable = FALSE;
 				return FALSE;
 			}
-			
-			str = g_strdup_printf ("%s%s", last_action->action.delete.text, 
+
+			str = g_strdup_printf ("%s%s", last_action->action.delete.text,
 				undo_action->action.delete.text);
-			
+
 			g_free (last_action->action.delete.text);
-			last_action->action.delete.end += (undo_action->action.delete.end - 
+			last_action->action.delete.end += (undo_action->action.delete.end -
 							   undo_action->action.delete.start);
 			last_action->action.delete.text = str;
 		}
 		else
 		{
 			gchar *str;
-			
+
 			/* Deleted with the backspace key */
 			if ((g_utf8_get_char (undo_action->action.delete.text) != ' ') &&
 			    (g_utf8_get_char (undo_action->action.delete.text) != '\t') &&
@@ -980,9 +974,9 @@ gtk_source_undo_manager_merge_action (GtkSourceUndoManager 	*um,
 				return FALSE;
 			}
 
-			str = g_strdup_printf ("%s%s", undo_action->action.delete.text, 
+			str = g_strdup_printf ("%s%s", undo_action->action.delete.text,
 				last_action->action.delete.text);
-			
+
 			g_free (last_action->action.delete.text);
 			last_action->action.delete.start = undo_action->action.delete.start;
 			last_action->action.delete.text = str;
@@ -991,10 +985,10 @@ gtk_source_undo_manager_merge_action (GtkSourceUndoManager 	*um,
 	else if (undo_action->action_type == GTK_SOURCE_UNDO_ACTION_INSERT)
 	{
 		gchar* str;
-		
+
 #define I (last_action->action.insert.chars - 1)
-		
-		if ((undo_action->action.insert.pos != 
+
+		if ((undo_action->action.insert.pos !=
 		     	(last_action->action.insert.pos + last_action->action.insert.chars)) ||
 		    ((g_utf8_get_char (undo_action->action.insert.text) != ' ') &&
 		      (g_utf8_get_char (undo_action->action.insert.text) != '\t') &&
@@ -1006,9 +1000,9 @@ gtk_source_undo_manager_merge_action (GtkSourceUndoManager 	*um,
 			return FALSE;
 		}
 
-		str = g_strdup_printf ("%s%s", last_action->action.insert.text, 
+		str = g_strdup_printf ("%s%s", last_action->action.insert.text,
 				undo_action->action.insert.text);
-		
+
 		g_free (last_action->action.insert.text);
 		last_action->action.insert.length += undo_action->action.insert.length;
 		last_action->action.insert.text = str;
@@ -1018,7 +1012,7 @@ gtk_source_undo_manager_merge_action (GtkSourceUndoManager 	*um,
 	else
 		/* Unknown action inside undo merge encountered */
 		g_return_val_if_reached (TRUE);
-		
+
 	return TRUE;
 }
 
@@ -1036,7 +1030,7 @@ gtk_source_undo_manager_set_max_undo_levels (GtkSourceUndoManager	*um,
 				  	     gint			 max_undo_levels)
 {
 	gint old_levels;
-	
+
 	g_return_if_fail (um != NULL);
 	g_return_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um));
 
@@ -1045,7 +1039,7 @@ gtk_source_undo_manager_set_max_undo_levels (GtkSourceUndoManager	*um,
 
 	if (max_undo_levels < 1)
 		return;
-		
+
 	if (old_levels > max_undo_levels)
 	{
 		/* strip redo actions first */
@@ -1054,7 +1048,7 @@ gtk_source_undo_manager_set_max_undo_levels (GtkSourceUndoManager	*um,
 			gtk_source_undo_manager_free_first_n_actions (um, 1);
 			um->priv->next_redo--;
 		}
-		
+
 		/* now remove undo actions if necessary */
 		gtk_source_undo_manager_check_list_size (um);
 
@@ -1066,7 +1060,7 @@ gtk_source_undo_manager_set_max_undo_levels (GtkSourceUndoManager	*um,
 		}
 
 		if (um->priv->can_undo &&
-		    um->priv->next_redo >= (gint)(g_list_length (um->priv->actions) - 1))
+		    um->priv->next_redo >= (gint)um->priv->actions->len - 1)
 		{
 			um->priv->can_undo = FALSE;
 			g_signal_emit (G_OBJECT (um), undo_manager_signals [CAN_UNDO], 0, FALSE);
@@ -1079,23 +1073,19 @@ gtk_source_undo_manager_modified_changed_handler (GtkTextBuffer        *buffer,
                                             	  GtkSourceUndoManager *um)
 {
 	GtkSourceUndoAction *action;
-	GList *list;
-	
+	gint index;
+
 	g_return_if_fail (GTK_SOURCE_IS_UNDO_MANAGER (um));
 	g_return_if_fail (um->priv != NULL);
 
-	if (um->priv->actions == NULL)
+	if (um->priv->actions->len == 0)
 		return;
 
-	list = g_list_nth (um->priv->actions, um->priv->next_redo + 1);
-
-	if (list != NULL)
-		action = (GtkSourceUndoAction*) list->data;
-	else
-		action = NULL;
+	index = um->priv->next_redo + 1;
+	action = action_list_nth_data (um->priv->actions, index);
 
 	if (gtk_text_buffer_get_modified (buffer) == FALSE)
-	{	
+	{
 		if (action != NULL)
 			action->mergeable = FALSE;
 
@@ -1116,20 +1106,22 @@ gtk_source_undo_manager_modified_changed_handler (GtkTextBuffer        *buffer,
 
 		return;
 	}
-	
+
 	/* gtk_text_buffer_get_modified (buffer) == TRUE */
 
-	g_return_if_fail (um->priv->modified_action == NULL);
+// 	g_return_if_fail (um->priv->modified_action == NULL);
+	if (um->priv->modified_action != NULL)
+	{
+		g_message ("%s: oops", G_STRLOC);
+		return;
+	}
 
 	if (action->order_in_group > 1)
 		um->priv->modified_undoing_group  = TRUE;
 
 	while (action->order_in_group > 1)
 	{
-		list = g_list_next (list);
-		g_return_if_fail (list != NULL);
-
-		action = (GtkSourceUndoAction*) list->data;
+		action = action_list_nth_data (um->priv->actions, ++index);
 		g_return_if_fail (action != NULL);
 	}
 
